/*
 * lmet.c
 *
 * GKrellM LongRun Meter Plugin for FreeBSD
 * Copyright (c) 2002 Masaharu FUJITA
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * You may contact the author by:
 *      e-mail: m@fjts.org
 */

#include <gkrellm/gkrellm.h>
#include <sys/sysctl.h>

#define PLUGIN "Lmet"
#define PLUGIN_KEYWORD "lmet"
#define VERSION "0.0.1"
#define SETMODECMD "/usr/X11R6/libexec/gkrellm/gkrellongrunctl "

#define LONGRUN_MODE_MINFREQUENCY       0x00
#define LONGRUN_MODE_ECONOMY            0x01
#define LONGRUN_MODE_PERFORMANCE        0x02
#define LONGRUN_MODE_MAXFREQUENCY       0x03
#define LONGRUN_MODE_UNKNOWN            0x04
#define LONGRUN_MODE_MAX                0x04

#define LongRunDataTypeNum 4
typedef enum {
  LongRun,
  Frequency,
  Voltage,
  Percentage,
} LongRunDataType;

static gint data[LongRunDataTypeNum];

gchar *longrun_mode_label[] = { "Min.", "Eco.", "Per.", "Max." };

static int oid[CTL_MAXNAME + 2][4];
static size_t oid_len[] = { CTL_MAXNAME + 2, CTL_MAXNAME + 2,
			    CTL_MAXNAME + 2, CTL_MAXNAME + 2 };

gint style_id;
Monitor *monitor;
Panel *panel;
Krell *krell;
Decal *decal;
char *label;
gint label_type = LongRun;
GtkWidget *config_radio[LongRunDataTypeNum+1];

GtkItemFactory *lmet_factory;

static void set_longrun_mode(GtkWidget *widget, gpointer data)
{
  gchar *cmd;
  
  cmd = g_strdup_printf("%s %d", SETMODECMD, GPOINTER_TO_INT(data));
  system(cmd);
  
  g_free(cmd);
}

GtkItemFactoryEntry lmet_factory_entry[] = {
  {"/-", NULL, NULL, 0,"<Separator>"},
  {"/Mini-frequency", NULL, set_longrun_mode,
   LONGRUN_MODE_MINFREQUENCY, "<Item>"},
  {"/Economy", NULL, set_longrun_mode,
   LONGRUN_MODE_ECONOMY, "<Item>"},
  {"/Performance", NULL, set_longrun_mode,
   LONGRUN_MODE_PERFORMANCE, "<Item>"},
  {"/Max-frequency", NULL, set_longrun_mode,
   LONGRUN_MODE_MAXFREQUENCY, "<Item>"},
  {"/-", NULL, NULL, 0,"<Separator>"},
};

GtkItemFactory *lmet_menu_factory()
{
  int size;
  GtkItemFactory *factory;
  GtkAccelGroup *accel = gtk_accel_group_new();

#if ((VERSION_MAJOR>1)||(VERSION_MAJOR==1 && VERSION_MINOR==1 && VERSION_REV>3)||(VERSION_MAJOR==1 && VERSION_MINOR>1))
  gtk_window_add_accel_group(GTK_WINDOW(gkrellm_get_top_window()), accel);
#else
  gtk_window_add_accel_group(GTK_WINDOW(top_window), accel);
#endif
  
  factory = gtk_item_factory_new(GTK_TYPE_MENU, "<Main>", accel);
  size = sizeof(lmet_factory_entry)/sizeof(GtkItemFactoryEntry);
  gtk_item_factory_create_items(factory, size, lmet_factory_entry, NULL);
  
  return factory;
}

#ifndef HAVE_SYSCTLNAMETOMIB
static int sysctlnametomib(const char *name, int *mibp, size_t *sizep)
{
  static int oid[2] = { 0, 3 };
  int error;
  
  *sizep *= sizeof(int);
  error = sysctl(oid, 2, mibp, sizep, name, (size_t)strlen(name));
  *sizep /= sizeof(int);
  if (error < 0) 
    return (error);
  return (0);
}
#endif

static gint check_longrun()
{
  static char *name[] = {
    "hw.crusoe.longrun",
    "hw.crusoe.frequency",
    "hw.crusoe.voltage",
    "hw.crusoe.percentage"
  };
  LongRunDataType i;
  
  /* Check Longrun. And set oid and oid_len (for sysctl()). */
  for (i = LongRun; i <= Percentage; ++i) {
    if (sysctlnametomib((void *)name[i],
                        oid[i], &oid_len[i]) < 0)
      return 0;
  }
  
  return 1;
}

static void get_longrun_data()
{
  size_t size;
  LongRunDataType i;

  for (i = LongRun; i <= Percentage; ++i) {
    size = sizeof(data[i]);
    if (sysctl(oid[i], oid_len[i], &data[i], &size, NULL, 0) < 0)
      return;
  }
}

static gint expose_event(GtkWidget *widget, GdkEventExpose *event)
{
  if (widget == panel->drawing_area)
    gdk_draw_pixmap(widget->window,
		    widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		    panel->pixmap,
		    event->area.x, event->area.y,
		    event->area.x, event->area.y,
		    event->area.width, event->area.height);
  
  return FALSE;
}

static gint clicked_event(GtkWidget *widget, GdkEventButton *event)
{
  switch (event->button) {
  case 3:
    gtk_menu_popup(GTK_MENU(lmet_factory->widget),
		   NULL, NULL, NULL, NULL, event->button, event->time);
    break;
  }
  
  return TRUE;
}

static void set_label()
{
  switch (label_type) {
  case LongRun:
    label = g_strdup_printf("%s", longrun_mode_label[data[LongRun]]);
    break;
  case Frequency:
    label = g_strdup_printf("%dMHz", data[Frequency]);
    break;
  case Voltage:
    label = g_strdup_printf("%.2fV %s", data[Voltage]/1000.0, longrun_mode_label[data[LongRun]]);
    break;
  case Percentage:
    label = g_strdup_printf("%03d%%  %s", data[Percentage], longrun_mode_label[data[LongRun]]);
    break;
  default:
    label = strdup("");
  }
}

static void create_plugin(GtkWidget *vbox, gint first)
{
  GdkImlibImage *bg_image;
  GdkImlibImage *krell_image;
  Style *panel_style;
  Style *slider_style;

  panel_style = gkrellm_meter_style(style_id);
  bg_image = gkrellm_bg_meter_image(style_id);

  if (first)
    panel = gkrellm_panel_new0();

  set_label();
  panel->textstyle = gkrellm_meter_textstyle(style_id);

  decal = gkrellm_create_decal_text(panel, "M8", panel->textstyle,
				    gkrellm_meter_style(style_id),
				    -1, -1, -1 );

  gkrellm_draw_decal_text(panel, decal, label, -1 );
  gkrellm_configure_panel(panel, NULL, panel_style);
  gkrellm_create_panel(vbox,panel,bg_image);

  gkrellm_monitor_height_adjust(panel->h);

  slider_style = gkrellm_krell_slider_style();
  krell_image = gkrellm_krell_slider_image();
  krell = gkrellm_create_krell(panel,krell_image,slider_style);
  krell->y0 = (panel->h - krell->h_frame)/2;
  krell->full_scale = 100;

  if (first) {
    gtk_signal_connect(GTK_OBJECT(panel->drawing_area), "expose_event",
		       (GtkSignalFunc) expose_event,NULL);
    gtk_signal_connect(GTK_OBJECT(panel->drawing_area),
		       "button_press_event",
		       (GtkSignalFunc) clicked_event,NULL);
  }

  gkrellm_update_krell(panel,krell,data[Percentage]);
  gkrellm_draw_layers(panel);
}

static void update_plugin()
{
  if (GK.second_tick)
    return;

  get_longrun_data();
  set_label();

  krell->previous = 0;
  gkrellm_update_krell(panel,krell,data[Percentage]);
  gkrellm_draw_decal_text(panel, decal, label, -1);
  gkrellm_draw_layers(panel);

}

static void create_tab(GtkWidget *tab)
{
  GtkWidget *tabs;
  GtkWidget *vbox, *about;
  GSList *label_group = NULL;
  gint i;
  gchar *label_name[] = {
    "Show mode", "Show frequency", "Show voltage + mode", "Show percentage + mode", "None"
  };

  gchar *about_text = 
    g_strdup_printf("%s %s\n" \
		    "GKrellM LongRun Meter Plugin\n" \
		    "\n" \
		    "Copyright (C) 2002 Masaharu FUJITA\n" \
		    "m@fjts.org\n" \
		    "\n" \
		    "Released under the GNU General Public License\n",
		    PLUGIN, VERSION);
  
  tabs = gtk_notebook_new();
  gtk_notebook_set_tab_pos(GTK_NOTEBOOK(tabs), GTK_POS_TOP);
  gtk_box_pack_start(GTK_BOX(tab), tabs, TRUE, TRUE, 0);

  /* Label Type  */
  vbox = gkrellm_create_tab(tabs, "Label Type");

  for ( i = LongRun ; i <= LongRunDataTypeNum ; i++ ) {
    config_radio[i] = gtk_radio_button_new_with_label(label_group,
						      label_name[i]);
    label_group = gtk_radio_button_group(GTK_RADIO_BUTTON(config_radio[i]));

    if ( i == label_type )
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(config_radio[i]), TRUE);

    gtk_widget_show(config_radio[i]);
    gtk_box_pack_start(GTK_BOX(vbox), config_radio[i], TRUE, TRUE, 0);
  }

  /* About */
  vbox = gkrellm_create_tab(tabs, "About");
  about = gtk_label_new(about_text);
  gtk_box_pack_start(GTK_BOX (vbox), about, TRUE, TRUE, 0);
}

static void apply_config()
{
  gint i;

  for ( i = LongRun ; i <= LongRunDataTypeNum ; i++ ) {
    if ( GTK_TOGGLE_BUTTON(config_radio[i])->active ) {
      label_type = i;
      break;
    }
  }

  set_label();
}

static void save_config(FILE *f)
{
  fprintf(f, "%s label_type %d\n", PLUGIN_KEYWORD, label_type);
}

static void load_config(gchar *arg)
{
  gchar config[32], item[CFG_BUFSIZE];
  
  if ( sscanf(arg, "%s %[^\n]", config, item) != 2 )
    return;
  
  if ( strcmp(config, "label_type") == 0 )
    sscanf(item, "%d", &label_type);
}

static Monitor lmet =
{
        "LongRun Meter",      /* Name, for config tab. */
        0,                      /* Id, 0 if a plugin */
        create_plugin,          /* The create_plugin() function */
        update_plugin,          /* The update_plugin() function */

        create_tab,             /* Create_plugin_tab() config func*/
        apply_config,           /* The apply_plugin_config() function*/

        save_config,            /* The save_plugin_config() function */
        load_config,            /* The load_plugin_config() function */
        PLUGIN_KEYWORD,         /* Config Keyword */

        NULL,                   /* Undefined 2 */
        NULL,                   /* Undefined 1 */
        NULL,                   /* Undefined 0 */

        MON_APM,               /* Insert plugin before this monitor */
        NULL,                   /* Handle if a plugin, filled by GKrellM */
        NULL                    /* Path if a plugin, filled by GKrellM */
};

Monitor *init_plugin()
{
  if (!check_longrun())
    return NULL;
  get_longrun_data();
  
  lmet_factory = lmet_menu_factory();
  monitor = &lmet;
  style_id = gkrellm_add_meter_style(&lmet, PLUGIN_KEYWORD);

  return &lmet;
}
